#include <R.h>
#include <Rinternals.h>
#include <Rdefines.h>
#include "R_ext/Memory.h"
#include "R_ext/Boolean.h"
#include "Utils.h"

char **
asCStringArray(USER_OBJECT_ svec)
{
    char **els = NULL;
 
    int i, n;
 
    n = GET_LENGTH(svec);
    if(n > 0) {
	els = (char **) R_alloc(n+1, sizeof(char*));
	for(i = 0; i < n; i++) {
	    els[i] = CHAR_DEREF(STRING_ELT(svec, i));
	}
        els[n] = NULL;
    }

    return(els); 
}

USER_OBJECT_
asRLogical(Rboolean val)
{
  USER_OBJECT_ ans;
  ans = NEW_LOGICAL(1);
  LOGICAL_DATA(ans)[0] = val;

  return(ans);
}


USER_OBJECT_
asRInt(int val)
{
  USER_OBJECT_ ans;
  ans = NEW_INTEGER(1);
  INTEGER_DATA(ans)[0] = val;

  return(ans);
}

USER_OBJECT_
asRNumeric(double val)
{
  USER_OBJECT_ ans;
  ans = NEW_NUMERIC(1);
  NUMERIC_DATA(ans)[0] = val;

  return(ans);
}

USER_OBJECT_
asRCharacter(const char *val)
{
  USER_OBJECT_ ans;
  PROTECT(ans = NEW_CHARACTER(1));
  if(val)
      SET_STRING_ELT(ans, 0, COPY_TO_USER_STRING(val));
  UNPROTECT(1);

  return(ans);
}



/**

 The idea is that the `val' argument is either an integer
 or a string. If it is an integer, we check that it is a
 value in the array cValues[] array.
 If it is a string, we look through the two arrays of names
 (the real names from the header files and the local names/aliases
  from the .defs files from which the enumerations are built.)

  If the value is matched, the associated integer is returned.

  We should probably put the real name on the result
  and also a class. In that case, we will need the name of the
  enum type.

  Otherwise, 
 */
USER_OBJECT_ 
S_checkEnum(USER_OBJECT_ val, const char *const localNames[], const char *const realNames[], 
                          const int cValues[], int len, const char *const enumName)
{
    int i;
    USER_OBJECT_ names, ans = NULL_USER_OBJECT;
    if(IS_INTEGER(val) || IS_NUMERIC(val)) {
        int cval;
        if(IS_INTEGER(val))
	    cval = INTEGER_DATA(val)[0];
        else 
	    cval = NUMERIC_DATA(val)[0];
	for(i = 0; i < len ; i++) {
	    if(cValues[i] == cval) {
		ans = val;
		break;
	    }
	}
        if(i == len) {
          PROBLEM "Invalid enum value: %d", INTEGER_DATA(val)[0]
    	  ERROR;
	}
    } else {
	char *tmp = CHAR_DEREF(STRING_ELT(val, 0));
	for(i = 0; i < len; i++) {
	    if(strcmp(tmp, localNames[i]) == 0 || strcmp(tmp, realNames[i]) == 0) {
                PROTECT(ans = NEW_INTEGER(1));
		INTEGER_DATA(ans)[0] = cValues[i];
		break;
	    }
	}
	if(i == len) {
          PROBLEM "Invalid enum name: %s", tmp
    	  ERROR;
	}
   }

    PROTECT(names = NEW_CHARACTER(1));
    SET_STRING_ELT(names, 0, COPY_TO_USER_STRING(realNames[i]));
    SET_NAMES(ans, names);

    PROTECT(names = NEW_CHARACTER(2));   
    SET_STRING_ELT(names, 0, COPY_TO_USER_STRING(enumName));
    SET_STRING_ELT(names, 1, COPY_TO_USER_STRING("enum"));
    SET_CLASS(ans, names);

    UNPROTECT(ans == val ? 2 : 3);

    return(ans);
}


/**
  val can be an integer or a character vector.
  If it is a character vector, we build up the actual value by 
  OR'ing the values in cValues[] by matching the elements in the vector
  with the elements in localNames or realNames.
 */

USER_OBJECT_ 
S_checkFlag(USER_OBJECT_ val, const char *const localNames[], const char *const realNames[], 
                          const int cValues[], int len, const char *const flagName)
{
    int i, j, n;
    USER_OBJECT_ ans, names;
    int numProtects = 0;
    n = GET_LENGTH(val);

    if(IS_INTEGER(val)) {
	for(i = 0; i < n ; i++) {
	    for(j = 0; j < len; j++) {
		if(!(cValues[j] | INTEGER_DATA(val)[i]))
		   PROBLEM "incorrect flag value: %d", INTEGER_DATA(val)[0]
  		   ERROR;
	    }
	}
        ans = val;
    } else {
	int value = 0;
	for(i = 0; i < n; i++) {
	    char *tmp = CHAR_DEREF(STRING_ELT(val, i));
            for(j = 0; j < len; j++)
	    if(strcmp(tmp, localNames[j]) == 0 || strcmp(tmp, realNames[j]) == 0) {
		value = value | cValues[j];
                break;
	    }
            if(j == len) {
               PROBLEM "Invalid flag name: %s", tmp
       	       ERROR;
	    }
	}
        PROTECT(ans = NEW_INTEGER(1));
        numProtects = 1;
        INTEGER_DATA(ans)[0] = value;
   }

    PROTECT(names = NEW_CHARACTER(2));   
    SET_STRING_ELT(names, 0, COPY_TO_USER_STRING(flagName));
    SET_STRING_ELT(names, 1, COPY_TO_USER_STRING("flag"));
    SET_CLASS(ans, names);

    UNPROTECT(numProtects + 1);

    return(ans);
}

USER_OBJECT_
R_toRFlag(int value, const char *flagName)
{
    USER_OBJECT_ ans, names;
    PROTECT(ans = NEW_INTEGER(1));
    INTEGER_DATA(ans)[0] = value;

    PROTECT(names = NEW_CHARACTER(2));   
    SET_STRING_ELT(names, 0, COPY_TO_USER_STRING(flagName));
    SET_STRING_ELT(names, 1, COPY_TO_USER_STRING("flag"));
    SET_CLASS(ans, names);

    UNPROTECT(2);
    return(ans);
}

SEXP toRList(SEXP * Ans, char** names, int length){
  int i;
  SEXP answer,attr;
    PROTECT(answer = allocVector(VECSXP,length));
    PROTECT(attr = allocVector(VECSXP,length));
    for (i=0;i<length;i++){
      SET_VECTOR_ELT(answer,i,Ans[i]);
      SET_VECTOR_ELT(attr,i,charToSexp(names[i],0));
    }
    setAttrib(answer, R_NamesSymbol, attr);
    UNPROTECT(2);
    return answer;
}


USER_OBJECT_
toRPointer(void *val, const char *type)
{
    USER_OBJECT_ ans;

    PROTECT(ans = R_MakeExternalPtr(val, Rf_install(type), NULL_USER_OBJECT));
    if(type) {
	SET_CLASS(ans, asRCharacter(type));
    }
    UNPROTECT(1);
  
    return(ans);
}

SEXP voidToSexp(void * x,int unused){
  return(toRPointer(x,"Void Pointer"));
}

SEXP returnNil(){
  return(voidToSexp(NULL,0));
}

void *
getPtrValue(USER_OBJECT_ sval)
{
  if(sval == NULL_USER_OBJECT)
      return(NULL);

  return(R_ExternalPtrAddr(sval));
}

void * SexpTovoid(SEXP z){
  return(getPtrValue(z));
}

void * SexpToVtk(USER_OBJECT_ sval){
  if(sval == NULL_USER_OBJECT)
      return(NULL);

  return(R_ExternalPtrAddr(sval));
}

SEXP VtkToSexp(void *val, const char *type)
{
    USER_OBJECT_ ans;

    PROTECT(ans = R_MakeExternalPtr(val, Rf_install(type), NULL_USER_OBJECT));
    if(type) {
	SET_CLASS(ans, asRCharacter(type));
    }
    UNPROTECT(1);
  
    return(ans);
}


char *
SexpTochar(USER_OBJECT_ svec)
{
    //    int n = strlen(CHAR(STRING_ELT(AS_CHARACTER(svec),0)));
    char * els = CHAR(STRING_ELT(AS_CHARACTER(svec), 0));

    //    int i;

    //    n = GET_LENGTH(svec);
    //    printf("%i",n);
    //    if(n > 0) {
	//	els = (char *) R_alloc(n+1, sizeof(char));
    //	for(i = 0; i < n; i++) {
    //	    els[i] = *R_CHAR(STRING_ELT(svec, i));
//   /}
    //        els[n] = NULL;
//   //}

    return(els); 
}

SEXP charToSexp(char * val,int unused) {
  USER_OBJECT_ ans;
  PROTECT(ans = NEW_CHARACTER(1));
  if(val)
      SET_STRING_ELT(ans, 0, COPY_TO_USER_STRING(val));
  UNPROTECT(1);

  return(ans);
}

SEXP doubleToSexp(double * d, int dims){
  int i;
  SEXP ans;
  if (dims<1) {
    ans=toRPointer(d,"Double Pointer");
  }  else {
    PROTECT(ans = allocVector(REALSXP, dims));
    for (i=0;i<dims;i++){
      REAL(ans)[i]= d[i];
    }
    UNPROTECT(1);
  }
  return ans;
}


double * SexpTodouble(USER_OBJECT_ svec)
{
    double * els = NULL;
 
    int i, n;
 
    n = GET_LENGTH(svec);
    if(n > 0) {
	els = (double *) R_alloc(n, sizeof(double));
	for(i = 0; i < n; i++) {
	    els[i] = NUMERIC_DATA(AS_NUMERIC(svec))[i];
	}
    }

    return(els); 
}

int * SexpToint(USER_OBJECT_ svec)
{
    int * els = NULL;
 
       int i, n;
 
       n = GET_LENGTH(svec);
       if(n > 0) {
    	els = (int *) R_alloc(n, sizeof(int));
    	for(i = 0; i < n; i++) {
    	    els[i] = NUMERIC_DATA(AS_NUMERIC(svec))[i];
    	}
        }

    return(els); 
}

SEXP intToSexp(int * d, int dims){
  int i;
  SEXP ans;
  if (dims<1) {
     ans=toRPointer(d,"Int Pointer");
  }  else {
    PROTECT(ans = allocVector(INTSXP, dims));
    for (i=0;i<dims;i++){
      INTEGER(ans)[i]= d[i];
    }
    UNPROTECT(1);
  }
  return ans;
}

short int * SexpToshort_int(USER_OBJECT_ svec)
{
    short int * els = NULL;
 
       int i, n;
 
       n = GET_LENGTH(svec);
       if(n > 0) {
    	els = (short int *) R_alloc(n, sizeof(int));
    	for(i = 0; i < n; i++) {
    	    els[i] = NUMERIC_DATA(AS_NUMERIC(svec))[i];
    	}
        }

    return(els); 
}

SEXP short_intToSexp(short int * d, int dims){
  int i;
  SEXP ans;
  if (dims<1) {
     ans=toRPointer(d,"Int Pointer");
  }  else {
    PROTECT(ans = allocVector(INTSXP, dims));
    for (i=0;i<dims;i++){
      INTEGER(ans)[i]= d[i];
    }
    UNPROTECT(1);
  }
  return ans;
}


SEXP long_unsigned_intToSexp(long unsigned int * d, int dims){
  int i;
  SEXP ans;
  if (dims<1) {
     ans=toRPointer(d,"Long_Unsigned_Int Pointer");
  } else {  
    PROTECT(ans = allocVector(INTSXP, dims));
    for (i=0;i<dims;i++){
      INTEGER(ans)[i]= d[i];
    }
    UNPROTECT(1);
  }
  return ans;
}

SEXP long_intToSexp(long int * d, int dims){
  int i;
  SEXP ans;
  if (dims<1) {
     ans=toRPointer(d,"Long_Int Pointer");
  }  else {
    PROTECT(ans = allocVector(INTSXP, dims));
    for (i=0;i<dims;i++){
      INTEGER(ans)[i]= d[i];
    }
  
  UNPROTECT(1);
  }
  return ans;
}

SEXP floatToSexp(float * d, int dims){
  int i;
  SEXP ans;
  if (dims<1) {
    ans=toRPointer(d,"Float Pointer");
  }  else {
    PROTECT(ans = allocVector(REALSXP, dims));
    for (i=0;i<dims;i++){
      REAL(ans)[i]= d[i];
    }
    UNPROTECT(1);
  }
  return ans;
}


float * SexpTofloat(USER_OBJECT_ svec)
{
    float * els = NULL;
 
    int i, n;
 
    n = GET_LENGTH(svec);
    if(n > 0) {
	els = (float *) R_alloc(n, sizeof(float));
	for(i = 0; i < n; i++) {
	    els[i] = NUMERIC_DATA(AS_NUMERIC(svec))[i];
	}
    }

    return(els); 
}

long int * SexpTolong_int(USER_OBJECT_ svec)
{
    long int * els = NULL;
 
    int i, n;
 
    n = GET_LENGTH(svec);
    if(n > 0) {
	els = (int *) R_alloc(n, sizeof(long int));
	for(i = 0; i < n; i++) {
	    els[i] = NUMERIC_DATA(AS_NUMERIC(svec))[i];
	}
    }

    return(els); 
}

unsigned char * SexpTounsigned_char(USER_OBJECT_ svec)
{
    unsigned char  * els = NULL;
 
    int i, n;
 
    n = GET_LENGTH(svec);
    if(n > 0) {
	els = (int *) R_alloc(n, sizeof(unsigned char));
	for(i = 0; i < n; i++) {
	    els[i] = NUMERIC_DATA(AS_NUMERIC(svec))[i];
	}
    }

    return(els); 
}

long unsigned int * SexpTolong_unsigned_int(USER_OBJECT_ svec)
{
    long unsigned int  * els = NULL;
 
    int i, n;
 
    n = GET_LENGTH(svec);
    if(n > 0) {
	els = (long unsigned int *) R_alloc(n, sizeof(long unsigned int));
	for(i = 0; i < n; i++) {
	    els[i] = NUMERIC_DATA(AS_NUMERIC(svec))[i];
	}
    }

    return(els); 
}

unsigned int * SexpTounsigned_int(USER_OBJECT_ svec)
{
    unsigned int  * els = NULL;
 
    int i, n;
 
    n = GET_LENGTH(svec);
    if(n > 0) {
	els = (unsigned int *) R_alloc(n, sizeof(unsigned int));
	for(i = 0; i < n; i++) {
	    els[i] = NUMERIC_DATA(AS_NUMERIC(svec))[i];
	}
    }

    return(els); 
}

short unsigned int * SexpToshort_unsigned_int(USER_OBJECT_ svec)
{
    short unsigned int  * els = NULL;
 
    int i, n;
 
    n = GET_LENGTH(svec);
    if(n > 0) {
	els = (short unsigned int *) R_alloc(n, sizeof(short unsigned int));
	for(i = 0; i < n; i++) {
	    els[i] = NUMERIC_DATA(AS_NUMERIC(svec))[i];
	}
    }

    return(els); 
}

SEXP unsigned_charToSexp(unsigned char * d, int dims){
  int i;
  SEXP ans;
  if (dims<1) {
     ans=toRPointer(d,"unsigned_char Pointer");
  } else {  
    PROTECT(ans = allocVector(INTSXP, dims));
    for (i=0;i<dims;i++){
      INTEGER(ans)[i]= d[i];
    }
    UNPROTECT(1);
  }
  return ans;
}

SEXP short_unsigned_intToSexp(short unsigned int * d, int dims){
  int i;
  SEXP ans;
  if (dims<1) {
     ans=toRPointer(d,"short_unsigned_int Pointer");
  } else {  
    PROTECT(ans = allocVector(INTSXP, dims));
    for (i=0;i<dims;i++){
      INTEGER(ans)[i]= d[i];
    }
    UNPROTECT(1);
  }
  return ans;
}

SEXP unsigned_intToSexp(unsigned int * d, int dims){
  int i;
  SEXP ans;
  if (dims<1) {
     ans=toRPointer(d,"unsigned_int Pointer");
  } else {  
    PROTECT(ans = allocVector(INTSXP, dims));
    for (i=0;i<dims;i++){
      INTEGER(ans)[i]= d[i];
    }
    UNPROTECT(1);
  }
  return ans;
}


void RVoidFunc(void * arg){
  RVoidFuncArg * z = arg;
  eval(z->func, z->env);
}

void RVoidFuncArgDelete(void * arg){
  Free(arg);
}
